//
//  GameScene.swift
//  MyGame
//
//  Created by 杉本裕樹 on 2017/01/14.
//  Copyright © 2017年 cl. All rights reserved.
//

import SpriteKit
import GameplayKit

class GameScene: SKScene {
    let player = SKShapeNode(circleOfRadius: 10)
    var enemies = [SKShapeNode]()
    var timer: Timer?
    var prevTime: TimeInterval = 0
    var startTime: TimeInterval = 0
    var isGameFinished = false
    let playerAgent = GKAgent2D()
    let agentSystem = GKComponentSystem(componentClass: GKAgent2D.self)
    var enemyAgents = [GKAgent2D]()
    var obstacles = [GKCircleObstacle]()
    var sweeper = SKShapeNode(circleOfRadius: 10)
    var sweeperAgent = GKAgent2D()
    
    override func didMove(to view: SKView) {
        player.fillColor = UIColor(red: 0.93, green: 0.96, blue: 0.00, alpha: 1.0)
        player.physicsBody = SKPhysicsBody(circleOfRadius: 10)
        addChild(player)
        
        sweeper.fillColor = UIColor(red: 255, green: 0.65, blue: 0.02, alpha: 1.0)
        sweeper.physicsBody = SKPhysicsBody(circleOfRadius: 10)
        addChild(sweeper)
        
        sweeperAgent.maxAcceleration = 50
        sweeperAgent.maxSpeed = 80
        sweeperAgent.delegate = self
        agentSystem.addComponent(sweeperAgent)
        
        createObstacles()
        setCreateEnemyTimer()
        physicsWorld.gravity = CGVector()
    }
    
    func createObstacles() {
        guard let viewFrame = view?.frame else {
            return
        }
        
        while obstacles.count < 5 {
            let point = CGPoint(
                x: CGFloat(arc4random_uniform(UInt32(viewFrame.width))) - viewFrame.width / 2,
                y: CGFloat(arc4random_uniform(UInt32(viewFrame.height))) - viewFrame.height / 2)
            let radius = Float(arc4random_uniform(50) + 50)
            
            // 障害物かPlayerが衝突していたら設置しない
            let isObstacleOverlapped = obstacles.contains {
                let dx = (Float(point.x) - $0.position.x)
                let dy = (Float(point.y) - $0.position.y)
                if sqrt(dx*dx + dy*dy) < $0.radius + radius {
                    return true
                }
                return false
            }
            let dx = point.x - player.position.x
            let dy = point.y - player.position.y
            let isPlayerOverlapped = sqrt(dx*dx + dy*dy) < CGFloat(radius) + player.frame.width
            if isObstacleOverlapped || isPlayerOverlapped {
                continue
            }
            
            let obstacleNode = SKShapeNode(circleOfRadius: CGFloat(radius))
            obstacleNode.fillColor = UIColor(red: 0.7, green: 0.7, blue: 0.7, alpha: 1.0)
            obstacleNode.position = point
            obstacleNode.physicsBody = SKPhysicsBody(circleOfRadius: CGFloat(radius))
            obstacleNode.physicsBody?.pinned = true
            addChild(obstacleNode)
            
            let obstacle = GKCircleObstacle(radius: radius)
            obstacle.position = float2(x: Float(point.x), y: Float(point.y))
            obstacles.append(obstacle)
        }
    }
    
    override func touchesEnded(_ touches: Set<UITouch>, with event: UIEvent?) {
        touches.forEach {
            let point = $0.location(in: self)
            player.removeAllActions()
            
            let path = CGMutablePath()
            path.move(to: CGPoint())
            path.addLine(to: CGPoint(x: point.x - player.position.x, y: point.y - player.position.y))
            player.run(SKAction.follow(path, speed: 50.0))
        }
    }
    
    func setCreateEnemyTimer() {
        timer?.invalidate()
        timer = Timer.scheduledTimer(timeInterval: 5.0, target: self, selector: #selector(GameScene.createEnemy), userInfo: nil, repeats: true)
        timer?.fire()
    }
    
    func createEnemy() {
        guard let viewFrame = view?.frame else {
            return
        }
        let xDistribution = GKShuffledDistribution(lowestValue: -Int(viewFrame.width / 2), highestValue: Int(viewFrame.width / 2))
        let yDistribution = GKShuffledDistribution(lowestValue: -Int(viewFrame.height / 2), highestValue: Int(viewFrame.height / 2))
        
        let enemy = SKShapeNode(circleOfRadius: 10)
        enemy.position.x = CGFloat(xDistribution.nextInt())
        enemy.position.y = CGFloat(yDistribution.nextInt())
        enemy.fillColor = UIColor(red: 0.94, green: 0.14, blue: 0.08, alpha: 1.0)
        enemy.physicsBody = SKPhysicsBody(circleOfRadius: enemy.frame.width / 2)
        addChild(enemy)
        enemies.append(enemy)
        
        let anemyAgent = GKAgent2D()
        anemyAgent.maxAcceleration = 200
        anemyAgent.maxSpeed = 70
        anemyAgent.position = vector_float2(x: Float(enemy.position.x), y: Float(enemy.position.y))
        anemyAgent.delegate = self
        anemyAgent.behavior = GKBehavior(goals: [
            GKGoal(toSeekAgent: playerAgent),
            GKGoal(toAvoid: obstacles, maxPredictionTime: 10),
            GKGoal(toFleeAgent: sweeperAgent)
            ], andWeights: [NSNumber(value: 5), NSNumber(value: 50), NSNumber(value: 1)])
        agentSystem.addComponent(anemyAgent)
        enemyAgents.append(anemyAgent)
        
        let goals = enemyAgents.map { GKGoal(toSeekAgent: $0) }
        sweeperAgent.behavior = GKBehavior(goals: goals + [GKGoal(toAvoid: obstacles, maxPredictionTime: 10)])
    }
    
    override func update(_ currentTime: TimeInterval) {
        if prevTime == 0 {
            prevTime = currentTime
            startTime = currentTime
        }
        agentSystem.update(deltaTime: currentTime - prevTime)
        playerAgent.position = vector_float2(x: Float(player.position.x), y: Float(player.position.y))
        
        if !isGameFinished {
            for enemy in enemies {
                let dx = enemy.position.x - player.position.x
                let dy = enemy.position.y - player.position.y
                if sqrt(dx*dx + dy*dy) < player.frame.width / 2 + enemy.frame.width / 2 {
                    isGameFinished = true
                    timer?.invalidate()
                    let label = SKLabelNode(text: "記録:\(Int(currentTime - startTime))秒")
                    label.fontSize = 80
                    label.position = CGPoint(x: 0, y: -100)
                    addChild(label)
                    break
                }
            }
            
            for (i, enemy) in enemies.enumerated() {
                let dx = enemy.position.x - sweeper.position.x
                let dy = enemy.position.y - sweeper.position.y
                if sqrt(dx*dx + dy*dy) < enemy.frame.width / 2 + sweeper.frame.width / 2 + 5 {
                    enemy.removeFromParent()
                    enemies.remove(at: i)
                    agentSystem.removeComponent(enemyAgents[i])
                    enemyAgents.remove(at: i)
                    let goals = enemyAgents.map { GKGoal(toSeekAgent: $0) }
                    sweeperAgent.behavior = GKBehavior(goals: goals + [GKGoal(toAvoid: obstacles, maxPredictionTime: 10)])
                    break
                }
            }
        }
        
        prevTime = currentTime
    }
}

extension GameScene: GKAgentDelegate {
    func agentDidUpdate(_ agent: GKAgent) {
        if let agent = agent as? GKAgent2D, let index = enemyAgents.index(where: { $0 == agent }) {
            let enemy = enemies[index]
            enemy.position = CGPoint(x: CGFloat(agent.position.x), y: CGFloat(agent.position.y))
        }
        if let agent = agent as? GKAgent2D, agent == sweeperAgent {
            sweeper.position = CGPoint(x: CGFloat(agent.position.x), y: CGFloat(agent.position.y))
        }
    }
}
